/*
 * Copyright (C) 2015 Alibaba Group Holding Limited.
 * ALL RIGHTS RESERVED.
 */

package com.yunos.tv.karaoke.demo;

import android.app.Activity;
import android.os.Bundle;
import android.os.Environment;
import android.os.Handler;
import android.os.Message;
import android.util.Log;
import android.view.View;
import android.widget.*;
import com.yunos.tv.karaoke.DeviceStatusListener;
import com.yunos.tv.karaoke.KaraokeController;
import com.yunos.tv.karaoke.ReverbControlNative;

import java.util.List;

public class ReverbDemo extends Activity implements AdapterView.OnItemSelectedListener, DeviceStatusListener, Handler.Callback {
    /**
     * Called when the activity is first created.
     */

    public static final String TAG = "ReverbDemo";

    private Button mStartButton;
    private Button mStopButton;
    private Button mPauseButton;
    private SeekBar mPlayProgress;
    private SeekBar mMusicVol;
    private SeekBar mVoiceVol;
    private SeekBar mReverbLevel;
    private CheckBox mUsbCheckBox;
    private CheckBox mRcCheckBox;
    private TextView mStatusTextView;
    private TextView mCurrentInputTextView;
    private TextView mCurrentStateTextView;
    private TextView mAvalibleOutput;
    private KPlayer mSimplePlayer;

    private AudioRecord mAudioDataListener;
    private KaraokeController mKaraokeController;

    private boolean mPaused;
    private int mCurrentSource;
    private static final String mSampleMp3File = Environment.getExternalStorageDirectory() + "/music_48.mp3";

    private Handler mMainHandler = new Handler(this);

    static final int MSG_STATE_CHANGE = 1;
    static final int MSG_STATE_ERROR = 2;

    @Override
    public void onCreate(Bundle savedInstanceState) {
        super.onCreate(savedInstanceState);
        setContentView(R.layout.main);

        initAudio();
        initPlayer();
        initView();
    }

    @Override
    protected void onStop() {
        super.onStop();

        mSimplePlayer.stop();
        mKaraokeController.stopReverb();

    }

    @Override
    protected void onDestroy() {
        super.onDestroy();
        mKaraokeController.closeDevice();
    }

    private  void initAudio() {

        // karaoke Controller is single instance, use get Instance to get it.
        // it's thread-safe.
        mKaraokeController = KaraokeController.getInstance(this);

        // Then,you should use this API check whether your device support karaoke or not.
        // if this return false, means this device don't supoprt this SDK.
        // you should use some other backup plan, like use system audio track, etc.
        Log.e("Demo", "is Supported " + mKaraokeController.isSupported(getApplicationContext()));


        // mKaraokeController require set the audio source and sink(play before create Mixer
        // USB means external device,
        // We also have type of device have Remote controller input device.

        // You'd better query which input device you have, if you don't have input device.
        // you can use muted device to play music only , or use system audio track.
        if (mKaraokeController.isExternalUsbMicAvalible()) {
            mCurrentSource = ReverbControlNative.AUDIO_SOURCE_USB;
        } else if (mKaraokeController.isRcMicAvalible()) {
            mCurrentSource = ReverbControlNative.AUDIO_SOURCE_RC;
        } else {
            // use system audio or use muted input source.
            mCurrentSource = ReverbControlNative.AUDIO_SOURCE_MUTED;
            // or use system audio, because wichout mic input, sdk does nothing.
        }

        mKaraokeController.setSource(mCurrentSource);


        // For output(sink) you can have a output type by this getAvalibleSinkType() function.
        // you can get a list.
        // you can let user choose which output they want , or just use some default config value.
        List<Integer> sinklist = mKaraokeController.getAvalibleSinkType();
        if (sinklist.size() == 1) {
            // the only output we have, use that one.
            Log.e("Demo", "set sink type: " +  sinklist.get(0));
            mKaraokeController.setSink(sinklist.get(0));
        } else {
            // better have a config or interactive UI let user choose which output they want.
            // AV means AV output port.
            // HDMI means HDMI, usually HDMI have more latency then AV port.
            mKaraokeController.setSink(ReverbControlNative.AUDIO_SINK_AV);
            //mKaraokeController.setSink(ReverbControlNative.AUDIO_SINK_HDMI);
        }

        // This call register device hot-plug callabck, in case user plugout Usb micphone.
        mKaraokeController.registerDeviceStatusListener(this);

        // This register controller state and error listener, will send back error and play state by this callback.
        mKaraokeController.registerControllerStateChangeListener(new ControllerStateListener());

        // This is the demo lib that use native API.
        mAudioDataListener = new AudioRecord();
        mAudioDataListener.init();
    }

    private void initView() {
        mStartButton = (Button) findViewById(R.id.start);
        mStopButton = (Button) findViewById(R.id.stop);
        mPauseButton = (Button)findViewById(R.id.pause_btn);
        mPlayProgress = (SeekBar) findViewById(R.id.play_progress);
        mMusicVol = (SeekBar)findViewById(R.id.music_vol_seekbar);
        mVoiceVol = (SeekBar)findViewById(R.id.micphone_vol_seekbark);
        mStatusTextView = (TextView) findViewById(R.id.play_status);
        mUsbCheckBox = (CheckBox)findViewById(R.id.usb_mic_check);
        mRcCheckBox = (CheckBox)findViewById(R.id.rc_mic_check);
        mCurrentInputTextView = (TextView)findViewById(R.id.current_input);
        mCurrentStateTextView = (TextView)findViewById(R.id.current_state);
        mReverbLevel = (SeekBar)findViewById(R.id.reverbSeekBar);
        mAvalibleOutput = (TextView)findViewById(R.id.avalible_out);
        String out = "";
        for (Integer o : mKaraokeController.getAvalibleSinkType()) {
            out = out.concat(" " + o);
        }
        Log.d(TAG, "Fetch sink list from controller:  " + out);

        mAvalibleOutput.setText(out);

        mMusicVol.setMax(100);
        mMusicVol.setProgress(80);  // default Volume is 80%
        mMusicVol.setOnSeekBarChangeListener(new VolumeBarChangeListener(VolumeBarChangeListener.MUSIC));
        mVoiceVol.setMax(100);
        mVoiceVol.setProgress(80);
        mVoiceVol.setOnSeekBarChangeListener(new VolumeBarChangeListener(VolumeBarChangeListener.VOICE));

        ArrayAdapter<CharSequence> adapter = ArrayAdapter.createFromResource(this,
                R.array.sound_effect_array, android.R.layout.simple_spinner_item);

        adapter.setDropDownViewResource(android.R.layout.simple_spinner_item);
        mUsbCheckBox.setChecked(mKaraokeController.isExternalUsbMicAvalible());
        mRcCheckBox.setChecked(mKaraokeController.isRcMicAvalible());

        mStartButton.setOnClickListener(new OnStartButtonClick());

        mStopButton.setOnClickListener(new OnStopButtonClick());

        mPauseButton.setOnClickListener(new OnPauseButtonClicked());
    }

    private void onStopReverb() {
        mSimplePlayer.stop();
        mKaraokeController.stopReverb();
        mKaraokeController.closeDevice();
    }

    private void setAudioInputLabel() {
        switch(mKaraokeController.getAudioSource()) {
            case ReverbControlNative.AUDIO_SOURCE_RC:
                mCurrentInputTextView.setText(getString(R.string.remote_controller_input));

                break;
            case ReverbControlNative.AUDIO_SOURCE_USB:
                mCurrentInputTextView.setText(getString(R.string.external_usb_input));
                break;
            case ReverbControlNative.AUDIO_SOURCE_MUTED:
                mCurrentInputTextView.setText(getString(R.string.muted_input));
                break;
        }
    }

    /**
     * Create a simple play to demo audio play and mixing.
     * The player is really simple, just for demo propose.
     */
    private void initPlayer() {
        PlayerEvents events = new PlayerEvents() {
            @Override public void onStop() {
                mPlayProgress.setProgress(0);
                mStatusTextView.setText("Playback Stopped.");
            }

            @Override public void onStart(String mime, int sampleRate, int channels, long duration) {
                Log.d(TAG, "onStart called: " + mime + " sampleRate:" + sampleRate + " channels:" + channels);
                mStatusTextView.setText("Playing content:" + mime + " " + sampleRate + "Hz " + (duration / 1000000) + "sec");
            }

            @Override public void onPlayUpdate(int percent, long currentms, long totalms) {
                mPlayProgress.setProgress(percent);
            }
            @Override public void onPlay() {
            }

            @Override public void onError() {
                mPlayProgress.setProgress(0);
                Toast.makeText(ReverbDemo.this, "Error!",  Toast.LENGTH_SHORT).show();
                mStatusTextView.setText("An error has been encountered");
            }
        };
        mSimplePlayer = new KPlayer(events);
    }


    @Override
    public void onInputDeviceStatusChanged(final int deviceType, boolean isPlugin) {
        // Callback when external device status have been changed.
        // This callback this not on Main thread, so call UI function in UI thread.

        if (deviceType == mCurrentSource && !isPlugin) {
            onStopReverb();
            runOnUiThread(new Runnable() {
                @Override
                public void run() {
                    Toast.makeText(ReverbDemo.this, "Play stop because current record device is plugin out.", Toast.LENGTH_LONG).show();
                }
            });
        }

        if (isPlugin) {
            runOnUiThread(new Runnable() {
                @Override
                public void run() {
                    Toast.makeText(ReverbDemo.this, "Input device " + deviceType + "isPlugin in, Can click start now." , Toast.LENGTH_LONG).show();
                }
            });
        }


    }

    @Override
    public boolean handleMessage(Message msg) {
        switch(msg.what) {
            case MSG_STATE_CHANGE:
                mCurrentStateTextView.setText((String) msg.obj);
                break;
            case MSG_STATE_ERROR:
                Toast.makeText(this, (String) " ERROR : " +  msg.obj, Toast.LENGTH_LONG).show();
        }
        return false;
    }

    private class VolumeBarChangeListener implements SeekBar.OnSeekBarChangeListener {
        public static final int MUSIC = 1;
        public static final int VOICE = 2;
        private final int mType;
        public VolumeBarChangeListener(int type) {
            super();
            mType = type;

        }

        // Control the sub-volume change.
        @Override
        public void onProgressChanged(SeekBar seekBar, int progress, boolean fromUser) {

            if (mType == MUSIC)
                mKaraokeController.changeMusicVolume(progress);

            else
                mKaraokeController.changeVoiceVolume(progress);
        }

        @Override
        public void onStartTrackingTouch(SeekBar seekBar) {

        }

        @Override
        public void onStopTrackingTouch(SeekBar seekBar) {

        }
    }

    // Control Reverberation(Reverb) level.
    private class ReverbLevelChangeListener implements SeekBar.OnSeekBarChangeListener {
        @Override
        public void onProgressChanged(SeekBar seekBar, int progress, boolean fromUser) {
            mKaraokeController.changeEffectParam(KaraokeController.kEffectParamReverb, progress);
        }

        @Override
        public void onStartTrackingTouch(SeekBar seekBar) {

        }

        @Override
        public void onStopTrackingTouch(SeekBar seekBar) {

        }
    }

    // The callback when controller state have changed.
    private class ControllerStateListener implements KaraokeController.StateListener {
        @Override
        public void onStateChanged(int newState) {


            if (mCurrentStateTextView != null) {
                Log.d(TAG, "state chagne to " + newState);
                Message msg = Message.obtain();
                msg.what = MSG_STATE_CHANGE;
                msg.obj = Integer.toString(newState);
                mMainHandler.sendMessage(msg);
            }
        }

        @Override
        public void onError(int errCode, boolean fatal) {
            if (mCurrentStateTextView != null) {
                Message msg = Message.obtain();
                msg.what = MSG_STATE_ERROR;
                msg.obj = Integer.toString(errCode);
                mMainHandler.sendMessage(msg);
                Log.d(TAG, "have error " + errCode);
            }

            if (fatal) {
                mKaraokeController.stopReverb();
                mSimplePlayer.stop();
            }
        }
    }

    private class OnStartButtonClick implements View.OnClickListener {

        @Override
        public void onClick(View v) {

            // create reverb also maybe failed.
            if (!mKaraokeController.createReverb()) {
               Log.e("Reverb", "error on create reverb.");
                return;
            }

            if (!mKaraokeController.startReverb()) {
                // you must check start Reverb return, it maybe failed!
                Log.e("Reverb", "error on start reverb, failed");
                return;
            }

            mSimplePlayer.setDataSource(mSampleMp3File);
            mSimplePlayer.play();
            setAudioInputLabel();

            mReverbLevel.setOnSeekBarChangeListener(new ReverbLevelChangeListener());

        }
    }

    private class OnStopButtonClick implements View.OnClickListener {
        @Override
        public void onClick(View v) {
            onStopReverb();
        }
    }

    private class OnPauseButtonClicked implements View.OnClickListener {
        @Override
        public void onClick(View v) {
            if (!mPaused) {
                mSimplePlayer.pause();
                mKaraokeController.pauseReverb();
                mPauseButton.setText("Resume");
                mPaused = true;
            } else {
                mSimplePlayer.play();
                mPauseButton.setText("Pause");
                mKaraokeController.resumeReverb();
                mPaused = false;
            }
        }
    }



    @Override
    public void onItemSelected(AdapterView<?> parent, View view, int position, long id) {

    }

    @Override
    public void onNothingSelected(AdapterView<?> parent) {

    }
}
